/**********************************************************BeginFile************
* Copyright:  (c) 2001 by Consul Risk Management B.V.  All Rights Reserved
* Time-stamp: <010809 Louis 16:00 list.c>
*
* Name:  list.c
* Purpose:
*   a general-purpose list
* Notes:
*   
* History:
*   010807 <Version> louis <ID>: Created.
***********************************************************EndFile*************/
#include <malloc.h>             /* for malloc(), free() */
#include <stdio.h>              /* for size_t */
#include <string.h>             /* for _strdup() */
#include "list.h"               /* for List, list_*() */

/* the data sizes that go with the recognized types */
static size_t list_type_size[] = {
  sizeof(void *), sizeof(unsigned int), sizeof(int), sizeof(float),
  sizeof(double), sizeof(char *), sizeof(long), sizeof(time_t)
};

#define DEFAULT_CAPACITY (16)

/**********************************************************BeginFunction********
* Name: list_create
* Syntax:
*  List list_create(List_type type, unsigned int grain) 
* Parameters:
*   List_type type     - specifies the type of data for the list
*                        (see list.h)
*   unsigned int grain - the number of elements by which to increase the
*                        capacity of the list if it fills up; 0 means use
*                        a default value.
* Return Values: the newly created List (a pointer), or NULL for an error.
* Purpose: To create a new list.
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
List list_create(List_type type, unsigned int grain) {
  List list;
  
  list = malloc(sizeof(_List));
  if (list) {
    list->type = type;
    list->size = 0;
    list->grain = grain? grain: DEFAULT_CAPACITY;
    list->capacity = list->grain;
    list->data.g = malloc(list->capacity*list_type_size[type]);
  }
  return list;
}

/**********************************************************BeginFunction********
* Name: list_append
* Syntax:
*    void list_append(List list, ...)
* Parameters:
*   List list - the list to append to
*   ...       - the single element to append, of the same type as the
*               one specified for the List when that was created.
* Return Values: 0 for success, non-zero for failure.
* Purpose: To append an element to an existing list.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
int list_append(List list, ...) {
  va_list ap;
  int result;

  va_start(ap, list);
  result = list_vinsert(list, list->size, ap);
  va_end(ap);
  return result;
}

/**********************************************************BeginFunction********
* Name: list_insert
* Syntax:
*  int list_insert(List list, unsigned int index, ...) 
* Parameters:
*   List list          - the list to insert into
*   unsigned int index - the index at which to insert (starts at 0)
*   ...                - the element to insert, of the same type as was
*                        specified when the list was created
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose:  To insert a new element into a list.  Elements in the list at
*   or beyond the specified index are moved one slot away from the beginning
*   of the list, and the new element is then inserted at the specified index.
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
int list_insert(List list, unsigned int index, ...) {
  va_list ap;
  int result;

  va_start(ap, index);
  result = list_vinsert(list, index, ap);
  va_end(ap);
  return result;
}

/**********************************************************BeginFunction********
* Name: list_swap
* Syntax:
*  int list_swap(List list, unsigned int index1, unsigned int index2) 
* Parameters:
*   List list           - the list
*   unsigned int index1 - the index of the first element (starts at 0)
*   unsigned int index2 - the index of the second element (starts at 0)
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: To swap two elements of a list.
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
int list_swap(List list, unsigned int index1, unsigned int index2) {
  double temp;                  /* no stack element needs more room than this */
  size_t size;
  
  if (index1 >= list->size || index2 >= list->size)
    return 1;
  size = list_type_size[list->type];
  memcpy(&temp, list->data.g + index1*size, size);
  memcpy(list->data.g + index1*size, list->data.g + index2*size, size);
  memcpy(list->data.g + index2*size, &temp, size);
  return 0;
}

/**********************************************************BeginFunction********
* Name: list_vinsert
* Syntax:
*  int list_vinsert(List list, unsigned int index, va_list ap) 
* Parameters:
*   List list          - the list to insert into
*   unsigned int index - the index at which to insert (starts at 0)
*   va_list ap         - contains the element to insert, of the same type as
*                        was specified when the list was created
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose:   To insert a new element into a list.  Elements in the list at
*   or beyond the specified index are moved one slot away from the beginning
*   of the list, and the new element is then inserted at the specified index.
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
int list_vinsert(List list, unsigned int index, va_list ap) {
  size_t size;
  
  if (index > list->size)
    return 1;
  size = list_type_size[list->type];
  if (list->size == list->capacity) {
    list->capacity += list->grain;
    list->data.g = realloc(list->data.g, list->capacity*size);
  }
  if (index < list->size)       /* not at the end; make room */
    memmove(list->data.g + (index + 1)*size, list->data.g + index*size,
            size*(list->size - index));
  list->size++;
  switch (list->type) {
    case s_string:
      list->data.s[index] = NULL; /* for proper memory management */
      break;
  }
  return list_vset(list, index, ap);
}

/**********************************************************BeginFunction********
* Name: list_size
* Syntax:
*  int list_size(List list) 
* Parameters:
*   List list - the list
* Return Values: The number of elements in the list
* Purpose: To return the number of elements in the list
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
int list_size(List list) {
  return list->size;
}

/**********************************************************BeginFunction********
* Name: list_read
* Syntax:
*  int list_read(List list, unsigned int index, void *dst) 
* Parameters:
*   List list          - the list from which to read an element
*   unsigned int index - the index of the element to read (starts at 0)
*   void *dst          - a pointer to the location for the result
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: To read an element from a list, without modifying the list.
* Example:
*   long l;
*   list_read(list, 0, &l);
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
int list_read(List list, unsigned int index, void *dst) {
  if (index > list->size)       /* non-existent element */
    return 1;
  switch (list->type) {
    case s_pointer:
      *(void **) dst = list->data.v[index];
      break;
    case s_uint:
      *(unsigned int *) dst = list->data.ui[index];
      break;
    case s_int:
      *(int *) dst = list->data.i[index];
      break;
    case s_float:
      *(float *) dst = list->data.f[index];
      break;
    case s_double:
      *(double *) dst = list->data.d[index];
      break;
    case s_string:
      *(char **) dst = list->data.s[index];
      break;
    case s_long:
      *(long *) dst = list->data.l[index];
      break;
    case s_time:
      *(time_t *) dst = list->data.t[index];
      break;
    default:        /* unknown type */
      return 2;
  }
  return 0;                     /* OK */
}

/**********************************************************BeginFunction********
* Name: list_clear
* Syntax:
*  void list_clear(List list) 
* Parameters:
*   List list - the list to clear
* Return Values: none
* Purpose: To clear all elements from a list, without changing the type or
*   capacity of the list
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
void list_clear(List list) {
  int i;
  
  switch (list->type) {
    case s_string:
      for (i = 0; i < list->size; i++)
        free(list->data.s[i]);
      break;
  }
  list->size = 0;
}

/**********************************************************BeginFunction********
* Name: list_delete
* Syntax:
*  void list_delete(List list) 
* Parameters:
*   List list - the list to delete
* Return Values: none
* Purpose: To delete the list and release any associated allocated memory
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
void list_delete(List list) {
  list_clear(list);
  free(list->data.g);
  free(list);
}

/**********************************************************BeginFunction********
* Name: list_vset
* Syntax:
*  int list_vset(List list, unsigned int index, va_list ap) 
* Parameters:
*   List list          - the list to modify
*   unsigned int index - the index of the element to modify (starts at 0)
*   va_list ap         - the new value for the element
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: to modify one element of a list, without changing its type
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
int list_vset(List list, unsigned int index, va_list ap) {
  if (index > list->size)
    return 1;
  switch (list->type) {
    case s_pointer: {
      void *ptr;

      ptr = va_arg(ap, void *);
      list->data.v[index] = ptr;
      break;
    }
    case s_uint: {
      unsigned int i;

      i = va_arg(ap, unsigned int);
      list->data.ui[index] = i;
      break;
    }
    case s_int: {
      int i;

      i = va_arg(ap, int);
      list->data.i[index] = i;
      break;
    }
    case s_float: {
      float f;

      f = va_arg(ap, float);
      list->data.f[index] = f;
      break;
    }
    case s_double: {
      double d;

      d = va_arg(ap, double);
      list->data.d[index] = d;
      break;
    }
    case s_string: {
      char *s;

      s = va_arg(ap, char *);
      if (list->data.s[index])
        free(list->data.s[index]);
      list->data.s[index] = _strdup(s);
      break;
    }
    case s_long: {
      long l;

      l = va_arg(ap, long);
      list->data.l[index] = l;
      break;
    }
    case s_time: {
      time_t t;
      
      t = va_arg(ap, time_t);
      list->data.t[index] = t;
      break;
    }
    default:
      return 2;
  }
  return 0;
}

/**********************************************************BeginFunction********
* Name: list_set
* Syntax:
*  int list_set(List list, unsigned int index, ...) 
* Parameters:
*   List list          - the list to modify
*   unsigned int index - the index of the element to modify (starts at 0)
*   ...                - the new value for the element
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: to modify one element of a list.
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
int list_set(List list, unsigned int index, ...) {
  va_list ap;
  int result;

  va_start(ap, index);
  result = list_vset(list, index, ap);
  va_end(ap);
  return result;
}
