/**********************************************************BeginFile************
* Copyright:  (c) 2001 by Consul Risk Management B.V.  All Rights Reserved
* Time-stamp: <010809 Louis 15:49 list.h>
*
* Name:  list.h
* Purpose:
*   a generic-purpose list
* Notes:
*   
* History:
*   010807 <Version> louis <ID>: Created.
***********************************************************EndFile*************/
#ifndef _LIST_
#define _LIST_
#ifdef __cplusplus
extern "C" {
#endif

#include <stdarg.h>             /* for ... */
#include <time.h>               /* for time_t */
  
/* recognized data types to put in a List */
  typedef enum {
    s_pointer, s_uint, s_int, s_float, s_double, s_string, s_long, s_time
  } List_type;

/* the List guts */
  typedef struct {
    List_type type;
    unsigned int size;
    unsigned int grain;
    unsigned int capacity;
    union {
      char *g;
      void **v;
      unsigned int *ui;
      int *i;
      float *f;
      double *d;
      char **s;
      long *l;
      time_t *t;
    } data;
  } _List;
  
  typedef _List* List;
  
/**********************************************************BeginFunction********
* Name: list_create
* Syntax:
*    List list_create(List_type type, unsigned int grain)
* Parameters:
*   List_type type     - the type of elements in the list; must be an
*                        element of List_type.
*   unsigned int grain - the number of elements by which to increase
*                        the capacity of the list if it fills up; 0 means
*                        use a default value.
* Return Values: the newly created List (a pointer), or NULL for an error.
* Purpose: To create a new list.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  List list_create(List_type type, unsigned int grain);

/**********************************************************BeginFunction********
* Name: list_append
* Syntax:
*    int list_append(List list, ...)
* Parameters:
*   List list - the list to append to
*   ...       - the single element to append, of the same type as the
*               one specified for the List when that was created.
* Return Values: 0 for success, non-zero for failure.
* Purpose: To append an element to an existing list.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  int list_append(List list, ...);
  
/**********************************************************BeginFunction********
* Name: list_swap
* Syntax:
*    int list_swap(List list, unsigned int index1, unsigned int index2)
* Parameters:
*   List list           - the list
*   unsigned int index1 - the index of the first element (starts at 0)
*   unsigned int index2 - the index of the second element (starts at 0)
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: To swap two elements of a list.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  int list_swap(List list, unsigned int index1, unsigned int index2);

/**********************************************************BeginFunction********
* Name: list_insert
* Syntax:
*    int list_insert(List list, unsigned int index, ...)
* Parameters:
*   List list          - the list to insert into
*   unsigned int index - the index at which to insert (starts at 0)
*   ...                - the element to insert, of the same type as was
*                        specified when the list was created
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: To insert a new element into a list.  Elements in the list at
*   or beyond the specified index are moved one slot away from the beginning
*   of the list, and the new element is then inserted at the specified index.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  int list_insert(List list, unsigned int index, ...);

/**********************************************************BeginFunction********
* Name: list_vinsert
* Syntax:
*    int list_vinsert(List list, unsigned int index, va_list ap)
* Parameters:
*   List list          - the list to insert into
*   unsigned int index - the index at which to insert (starts at 0)
*   va_list ap         - contains the element to insert, of the same type as
*                        was specified when the list was created
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose:  To insert a new element into a list.  Elements in the list at
*   or beyond the specified index are moved one slot away from the beginning
*   of the list, and the new element is then inserted at the specified index.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  int list_vinsert(List list, unsigned int index, va_list ap);

/**********************************************************BeginFunction********
* Name: list_size
* Syntax:
*    int list_size(List list)
* Parameters:
*   List list - the list of which the size is returned.
* Return Values: the number of elements currently in the list.
* Purpose: To return the number of elements that are currently in the list.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  int list_size(List list);

/**********************************************************BeginFunction********
* Name: list_read
* Syntax:
*    int list_read(List list, unsigned int index, void *dst)
* Parameters:
*   List list          - the list from which to read an element
*   unsigned int index - the index of the element to read (starts at 0)
*   void *dst          - a pointer to a location for the result
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: To read an element from a list, without modifying the list.
* Example:
*   long l;
*   list_read(list, 0, &l);
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  int list_read(List list, unsigned int index, void *dst);

/**********************************************************BeginFunction********
* Name: list_clear
* Syntax:
*    void list_clear(List list)
* Parameters:
*   List list - the list to clear
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: To clear all elements from a list, without changing the type or
*   capacity of the list.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  void list_clear(List list);

/**********************************************************BeginFunction********
* Name: list_delete
* Syntax:
*    void list_delete(List list)
* Parameters:
*   List list - the list to delete
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: To delete the list and release any associated allocated memory.
*   
* History:
*    20010805 louis: Created.
***********************************************************EndFunction*********/
  void list_delete(List list);

/**********************************************************BeginFunction********
* Name: list_set
* Syntax:
*    int list_set(List list, unsigned int index, ...)
* Parameters:
*   List list          - the list to modify
*   unsigned int index - the index of the element to modify (starts at 0)
*   ...                - the new value for the element
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: to modify one element of a list.
*   
* History:
*    20010809 louis: Created.
***********************************************************EndFunction*********/
  int list_set(List list, unsigned int index, ...);

/**********************************************************BeginFunction********
* Name: list_vset
* Syntax:
*  int list_vset(List list, unsigned int index, va_list ap)
* Parameters:
*   List list          - the list to modify
*   unsigned int index - the index of the element to modify (starts at 0)
*   va_list ap         - contains the new value for the element
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: to modify one element of a list.
*   
* History:
*    20010809 louis: Created.
***********************************************************EndFunction*********/
int list_vset(List list, unsigned int index, va_list ap);
  
#ifdef __cplusplus
}
#endif
#endif /* _LIST_ */
