/**********************************************************BeginFile************
* Copyright:  (c) 2001 by Consul Risk Management B.V.  All Rights Reserved
* Time-stamp: <040512 louis 12:49 opicsact.c>
*
* Name:  opicsact.c
* Purpose:
*   generic actuator executable for OPICS
* Notes:
*   the data consist of a number of files with names
*     w<daynum><workstationid>.log
*     u<daynum><userid>.log
*     b<daynum>.bsy
*   the data in those files is *almost* always in ascending temporal
*   order, but unfortunately there are rare exceptions that we must deal with.
* History:
*   010802 <Version> Louis <ID>: Created.
***********************************************************EndFile*************/

#include <stdio.h>              /* for printf(), puts(), FILE, fopen(), */
                                /* fclose() */
#include <io.h>                 /* for findfirst(), findnext(), findclose() */
#include <string.h>             /* for strcpy(), strcat(), strlen(),
                                   strncmp() */
#include <direct.h>             /* for chdir(), getcwd() */
#include <malloc.h>             /* for malloc(), free() */
#include <limits.h>             /* for UINT_MAX */
#include <time.h>               /* for time_t, time(), timezone */
#include <sys/types.h>          /* for stat() */
#include <sys/stat.h>           /* for stat() */
#include <errno.h>              /* for errno */
#include "rs_merge.h"           /* for rs_merge(), initialize_rs_merge() */
#include "list.h"               /* for list_*() */
#include "properties.h"         /* for Properties */

#define REJECT_FILE (_A_SUBDIR)

#define BUFSIZE 256

/* #define OPICS6 */

static FILE *msgfp;
static int debuglevel = 0;

static int n_files_open = 0, max_n_files_open = 0;

void print_log(char *fmt, ...)
{
  va_list ap;

  if (msgfp) {
    va_start(ap, fmt);
    vprintf(fmt, ap);
    vfprintf(msgfp, fmt, ap);
    fflush(msgfp);
    va_end(ap);
  }
}

FILE *my_fopen(const char *filename, const char *access) {
  FILE *fp;

  fp = fopen(filename, access);
  if (debuglevel) {
    if (fp) {
      if (++n_files_open > max_n_files_open)
        max_n_files_open = n_files_open;
      print_log("opened file %s, mode %s; now %1d files open (max %1d)\n",
                filename, access, n_files_open, max_n_files_open);
    } else
      print_log("could not open file %s, mode %s; now %1d files open (max %1d)\n",
                filename, access, n_files_open, max_n_files_open);
  }
  return fp;
}

int my_fclose(FILE *fp) {
  int result;

  result = fclose(fp);
  if (debuglevel) {
    if (!result) {
      --n_files_open;
      print_log("closed a file; now %1d files open (max %1d)\n", n_files_open,
                max_n_files_open);
    } else
      print_log("could not close a file; now %1d files open (max %1d)",
                n_files_open, max_n_files_open);
  }
  return result;
}

int my_fgets(char *buf, int limit, FILE *fp) {
  int index = 0;
  
  while (!feof(fp) && index < limit) {
    buf[index++] = fgetc(fp);
    if (buf[index - 1] == EOF) {
      --index;
      break;
    }
    if (buf[index - 1] == '\n')
      break;
  }
  buf[index] = '\0';
  return index;
}

int my_rename(const char *old, const char *new) {
  int result;

  result = rename(old, new);
  if (result)
    print_log("Renaming of %s to %s failed: [%d] %s\n", old, new, errno,
              strerror(errno));
  return result;
}

int my_remove(const char *file) {
  int result;

  result = remove(file);
  if (result)
    print_log("Removing of %s failed: [%d] %s\n", file, errno, strerror(errno));
  return result;
}

#ifdef OPICS6
void transform_files(List list) /* 040512 0.0.35 louis EE0404001: quick fix */
{
  char buf[BUFSIZE], *filename;
  int nfiles, file;
  int result;
  FILE *fp;

  nfiles = list_size(list);
  for (file = 0; file < nfiles; file++) {
    struct stat statbuf;
    
    list_read(list, file, &filename);
    /* make a copy of the original file with the same name with .old
     * appended */
    /* if the intended target already exists, then we assume that
     * the current file is already transformed */
    _snprintf(buf, BUFSIZE, "%s.old", filename);
    if (stat(buf, &statbuf) == 0) /* the file already exists */
      print_log("File %s already exists, so I assume %s is already ASCII\n",
                buf, filename);
    else
    {
      _snprintf(buf, BUFSIZE, "copy %s %s.old", filename, filename);
      result = system(buf);
      if (result != 0)
      {
        print_log("Error %d (%d) occurred when trying to copy %s to %s.old\n"
                  "The file will not be transformed\n",
                  result, result == -1? errno: 0, filename, filename);
        continue;
      }
      else
        print_log("Copied %s to %s.old\n", filename, filename);
      /* now transform the original from Unicode to ASCII */
      _snprintf(buf, BUFSIZE, "recode -q -t :ascii-bs %s", filename);
      result = system(buf);
      if (result != 0)
      {
        print_log("Error %d (%d) occurred when trying to transform"
                  " %s to ASCII\n",
                  result, result == -1? errno: 0, filename);
        continue;
      }
      else
        print_log("Transformed %s to ASCII\n", filename);
    }
  }
}
#endif

#define fopen(a,b) my_fopen(a,b)
#define fclose(a) my_fclose(a)
#define fgets(a,b,c) my_fgets(a,b,c)
#define rename(a,b) my_rename(a,b)
#define remove(a) my_remove(a)

void print_size(char *filename) {
  FILE *fp;

  fp = fopen(filename, "r");
  if (fp) {
    fseek(fp, 0, SEEK_END);
    print_log("File %s contains %1d bytes.\n", filename, ftell(fp));
    fclose(fp);
  } else
    print_log("Could not open file %s to determine its size.\n", filename);
}

void list_regular_files(int numpat, char *patterns[], List list)
/* seeks regular files matching the <numpat> DOS file name patterns
   in <patterns> and appends their names to the <list> (which must
   have type s_string). */
{
  char buf[BUFSIZE];
  int i, size;
  struct _finddata_t filedata;
  long handle;

  _getcwd(buf, BUFSIZE);
  if (debuglevel) {
    print_log("seeking regular files matching ");
    for (i = 0; i < numpat; i++)
      print_log("%s ", patterns[i]);
    print_log(" in directory %s\n", buf);
  }
  size = strlen(buf);
  buf[size++] = '\\';
  buf[size] = '\0';
  for (i = 0; i < numpat; i++) {
    handle = _findfirst(patterns[i], &filedata);
    if (handle != -1L) {        /* we found at least one */
      do
        if (size + strlen(filedata.name) + 1 < BUFSIZE) {
          strcpy(buf + size, filedata.name);
          if (debuglevel)
            print_log("found %s\n", filedata.name);
          list_append(list, buf);
        } else {
          buf[size] = '\0';
          print_log("file name too long in list_regular_files: %s\\%s\n",
                    buf, filedata.name);
        }
      while (!_findnext(handle, &filedata));
      _findclose(handle);
    } else if (debuglevel)
      print_log("found no files for pattern %s\n", patterns[i]);
  }
}

static struct _finddata_t filedata;
char *omf_lastfile(void)
{
 return filedata.name;
}

/**********************************************************BeginFunction********
* Name: open_many_files
* Syntax:
*   open_many_files(numpat, patterns, maxnumfile, fps)
* Parameters:
*   int numpat       - the number of DOS file name patterns in <patterns>,
*                      or zero, which triggers clean-up of the currently
*                      active search.
*   char *patterns[] - DOS file name patterns to match files names to.
*   int maxnumfile   - the maximum number of files to open.
*   FILE **fps       - preallocated room for <maxnumfile> file pointers.
* Return Values: the number of opened files.  If this is assigned to variable
*  n, then fps[0] through fps[n - 1] point to the opened files.  If no (more)
*  files were found, then 0 is returned.
* Purpose: To open up to <maxnumfile> file names in the current directory
*  whose names fit any of the <numpat> <patterns>, and return associated
*  file pointers.
* Comments:
*  Subsequent calls to open_many_files() continue where the previous one left
*  off, unless the previous call exhausted the current patterns, or
*  open_many_files() is called with <numpat> equal to 0.  If open_many_files()
*  returns zero, then proper clean-up has been done.  To break off a search
*  when open_many_files() did not yet return zero, call open_many_files()
*  with <numpat> equal to 0.  You are always responsible for closing the
*  files opened by open_many_files().
*
* Typical use:
*   while (n = open_many_files(numpat, patterns, maxnumfile, fps)) {
*     for (i = 0; i < n; i++) {
*       treat_file(fps[i]);
*       fclose(fps[i]);
*     }
*   }
*
* History:
*    Created 20010807 louis.
***********************************************************EndFunction*********/
int open_many_files(int numpat, char *patterns[], int maxnumfile, FILE **fps)
{
  static long handle = -1L;
  static int kind = 0, patindex = 0;
  int fpindex = 0;

  if (!numpat) {
    if (handle != -1L) {
      _findclose(handle);
      handle = -1L;
    }
    kind = patindex = 0;
    return 0;
  }

  while (1)
    switch (kind) {
      case 0:                   /* start on a new pattern */
        if (patindex == numpat) { /* we get here at the end of having */
                                  /* gone through a full set of files */
          patindex++;
          if (handle != -1L) {
            _findclose(handle);
            handle = -1L;
          }
          return fpindex;
        } else if (patindex > numpat) {
          /* the previous call to this routine exhausted the files, but */
          /* had to return the number of files in the last batch. */
          /* this time we indicate that we're done with this set of */
          /* files. */
          patindex = 0;         /* to start anew next time */
          return 0;             /* we're done */
        } else {
          handle = _findfirst(patterns[patindex], &filedata);
          if (handle == -1L) {
            patindex++;
            kind = 0;             /* to next pattern */
          } else 
            kind = 1;             /* to treatment of first file */
        }
        break;
      case 1:                   /* treat first file for new pattern */
        kind++;                 /* to remaining files */
        if ((filedata.attrib & REJECT_FILE) == 0) { /* only normal files */
          fps[fpindex] = fopen(filedata.name, "r");
          if (fps[fpindex]) {
            if (++fpindex == maxnumfile)
              return fpindex;
          } else
            print_log("Could not open file %s for reading\n", filedata.name);
        }
        break;
      case 2:                   /* treat the remaining files */
        while (_findnext(handle, &filedata) == 0) {
          if ((filedata.attrib & REJECT_FILE) == 0) { /* only normal files */
            fps[fpindex] = fopen(filedata.name, "r");
            if (fps[fpindex]) {
              if (++fpindex == maxnumfile)
                return fpindex;
            } else
              print_log("Could not open file %s for reading\n", filedata.name);
          }
        }
        _findclose(handle);
        handle = -1L;
        kind = 0;
        patindex++;
        break;
    }
}

typedef unsigned int uint;

/* to compare two unsigned integer keys and return -1, 0, or +1 depending
   on whether the first key is less than, equal to, or greater than
   the second one.  For use with rs_merge(). */
int keycmp(void *key1, void *key2) {
  uint w1, w2;

  w1 = *(uint *) key1;
  w2 = *(uint *) key2;

  return (w1 < w2)? -1: ((w1 > w2)? +1: 0);
}

/**********************************************************BeginFunction********
* Name: sort_files
* Syntax:
*  void sort_files(List list, char *auxprefix)
* Parameters:
*   List list       - a list of OPICS log files that must each be sorted into
*                     ascending temporal order
*   char *auxprefix - a prefix for temporary auxilliary files
* Return Values: none
* Purpose: to check if the contents of OPICS log files are in ascending
*   temporal order.  If any file is found whose contents are not completely
*   in ascending temporal order, then an auxilliary file with prefix <auxprefix>
*   is generated in which the contents of the file are stored in properly
*   sorted order.  The first auxilliary file gets _0 appended to <auxprefix>,
*   the next one _1, etcetera.  The names of the auxilliary files are
*   substituted for the names of their unsorted cousins in <list>.
*   
* History:
*   010809 x.y.z Louis <ID>: Created.
***********************************************************EndFunction*********/
void sort_files(List list, char *auxprefix)
{
  int nfiles, i, file;
  List timelist, ftellist;
  char *filename;
  FILE *fp, *outfp;
  char buf[BUFSIZE], *pbuf;
  time_t time, time0;
  struct tm bt;
  unsigned int bad, naux = 0;

  nfiles = list_size(list);
  timelist = list_create(s_time, 128);
  ftellist = list_create(s_long, 128);
  for (file = 0; file < nfiles; file++) {
    list_read(list, file, &filename);
    if (fp = fopen(filename, "r")) {
      long linesize, tell = 0;
        
      time = 0;
      bad = 0;
      while (fgets(buf, BUFSIZE, fp)) {
        pbuf = strchr(buf, '\x02');
        if (pbuf) {             /* a new record */
          /* incomplete records may occur; we seek the last ^B on the
             line and accept that as the start of the current record;
             any stuff before it is discarded. TODO: accept incomplete
             records if they're not too short */
          char *p;
            
          while (p = strchr(pbuf + 1, '\x02'))
            pbuf = p;
          time0 = time;
          bt.tm_year = 70;
          bt.tm_mon = 0;
          bt.tm_mday = atol(pbuf + 10) - 25568; /* # days since 1970-1-1 */
          bt.tm_hour = atol(pbuf + 16);
          bt.tm_min = atol(pbuf + 19);
          bt.tm_sec = atol(pbuf + 22);
          bt.tm_isdst = -1;     /* let the system determine DST */
          time = mktime(&bt);
        }
        /* else it is a second or further line from the previous
           record; we treat it as a separate record with the same time
           as the previous record; through careful insertion it should
           end up right after the previous record */
        if (time < time0) {     /* found a temporal discrepancy! */
          int i;
          unsigned int prevtime;
            
          bad++;            /* count it */
          for (i = list_size(timelist) - 1; i >= 0; i--) {
            /* seek proper place for it */
            list_read(timelist, i, &prevtime);
            if (prevtime < time)
              break;
          } /* end of for (i = line, ...) */
          list_insert(timelist, i + 1, time);
          list_insert(ftellist, i + 1, tell);
          time = time0;
        } else {
          list_append(timelist, time);
          list_append(ftellist, tell);
        } /* end of if (time < time0) */
        tell = ftell(fp);       /* current file position */
      } /* end of while (fgets()...) */
      if (bad) {                /* we did find some discrepancies */
        /* we now have, in linelist, a list of line numbers that will
           yield a time-sorted file */
        int i;
        char *oldname, *newname;

        sprintf(buf, "%s_%1d", auxprefix, naux);
        naux++;
        
        if (debuglevel)
          print_log("Opening file %s for writing\n", buf); /* DEBUG */
        outfp = fopen(buf, "w"); /* this will receive the sorted data */
        if (!outfp) {
          print_log("Could not open file %s for writing\n", buf);
          exit(1);
        }
        /* we must replace the original file name in the list of file
           names with the name of the auxilliary file, so that it and
           not the unsorted file gets further treatment */
        list_set(list, file, buf);
        for (i = 0; i < list_size(ftellist); i++) {
          char *o;
          
          list_read(ftellist, i, &tell);
          fseek(fp, tell, SEEK_SET);
          fgets(buf, BUFSIZE, fp); /* get the appropriate line */
          pbuf = strchr(buf, '\x02');
          if (pbuf) {           /* strip off incomplete records */
            char *p;
            while (p = strchr(pbuf + 1, '\x02'))
              pbuf = p;
          } else
            pbuf = buf;
          o = pbuf + strlen(pbuf) - 1;
          fputs(pbuf, outfp);   /* and write to the auxilliary file */
          if (*o != '\n') /* we found an explicit null byte in the data */
            fputc('\n', outfp);
        } /* end of for (i = 0; ...) */
        fclose(outfp);
      } /* end of if (bad) */
      fclose(fp);
      list_clear(timelist);
      list_clear(ftellist);
    } /* end of if (fp = fopen(...)) */
  } /* end of for (i = 0; ...) */
  list_delete(timelist);
  list_delete(ftellist);
}

/**********************************************************BeginFunction********
* Name: remove_files
* Syntax:
*  void remove_files(char *pattern)
* Parameters:
*   char *pattern - a DOS file name pattern
* Return Values: none.
* Purpose: Removes files that match the file name pattern.
*   
* History:
*    Created.
***********************************************************EndFunction*********/
void remove_files(char *pattern)
{
  long handle;
  struct _finddata_t filedata;

  handle = _findfirst(pattern, &filedata);
  do {
    if (debuglevel)
      print_log("Removing %s\n", filedata.name);
    remove(filedata.name);      /* remove() sometimes returns non-zero even
                                   if the remove completed succesfully. */
  } while (!_findnext(handle, &filedata));
  _findclose(handle);
}

#define FILEBATCH _NSTREAM_     /* max number of files to open at a time */
                                /* _NSTREAM_ is defined in stdio.h */

void usage(void)
{
  puts("Usage: opicsact\n\
   <outfile>      - the output file\n\
   (empty)\n\
   (empty)\n\
   (empty)\n\
   <datadir>      - the directory containing the OPICS log files\n\
   <messagefile>  - a file for progress and error messages (empty = no)\n\
   (empty) or \"debug\" - extra debugging info in message file \n\
   (empty)\n\
   (empty)\n\
   (empty)\n\
   (empty)\n\
   (empty)\n\
   <propsfile>    - a properties file name to remember the time\n\
   <starttime>    - the start of the time range to collect (UT, inclusive)\n\
   <stoptime>     - the end of the time range to collect (UT, exclusive)\n\n\
Short form:\n\
opicsact -<outfile> <datadir> <messagefile> <propsfile> <maxopenfiles> <starttime> <stoptime>\n\
 (with implicit debug messages)\n\
Only the arguments up to and including <messagefile> are required.\n\
Start and stop times: YYYY-MM-DD[-HH[:mm[:SS]]]");
  exit(1);
}

/**********************************************************BeginFunction********
* Name: main
* Syntax:
*  int main(int argc, char *argv[])
* Parameters:
*   int argc     - the number of arguments
*   char *argv[] - the arguments:
*     argv[0]    - the name of the program
*     argv[1]    - the name of the target file
*     argv[2-4]  - empty
*     argv[5]    - the name of the directory containing the per-day
*                  subdirectories with OPICS files
*     argv[6]    - file name for error messages (empty = no)
*     argv[7]    - if it starts with a d (e.g., debug), then extra debug
*                  messages are output to the log file
*     argv[8]    - the maximum number of files that may be open at any one time
*     argv[9-12] - empty
*     argv[13]   - the name of a properties file
*     argv[14]   - (for expert use only) the start of the time range
*                  for which data should be collected (inclusive)
*     argv[15]   - (for expert use only) the end of the time range
*                  for which data should be collected (exclusive)
*
* NOTE: to facilitate manual operation, a different argument list is assumed
*   if argv[1] begins with a minus sign: in that case, the empty arguments
*   (and the minus sign at the beginning of argv[1]) are omitted:
*     argv[0]    - the name of the program
*     argv[1]    - the name of the target file (starts with a -)
*     argv[2]    - the name of the directory containing the OPICS files
*     argv[3]    - file name for error messages (empty = no)
*     argv[4]    - the name of a properties file
*     argv[5]    - the maximum number of files that may be open at any one time
*     argv[6]    - the start of the time range for which data should
*                  be collected (inclusive)
*     argv[7]    - the end of the time range for which data should
*                  be collected (exclusive)
*
* The time range boundaries should be specified in the following form:
* YYYY-MM-DD_[HH[:MM[:SS]]] where the [] indicate optional parts.
* YYYY = the year number, MM = the month number (January = 1), DD =
* the day number, HH = the hour, MM = the minute, SS = the second.
* The time is interpreted as UT.
*
* Return Values: success = 0 failure = ERROR_CODE>0 
* Purpose: 
*   To generate chunk logs from OPICS log files.  All log files with
*   names fitting [0-9]+\\[uw]*.log and [0-9]+\\b*.bsy in the target
*   directory are taken to be OPICS log files, and a single chunk log
*   file is created with the records from the OPICS files that fall
*   within the target time range.  The target time range is determined
*   as follows:
*
*   (1) if a valid time range start is specified on the command line,
*   then that value is used.  Otherwise, if the properties file exists
*   and contains a valid "lastrun" property, then its value is used.
*   Otherwise, 0 is used.
*   (2) if a valid time range end is specified on the command line,
*   then that value is used.  Otherwise, the current time is used.
*
* The "lastrun" property in the properties file is updated when the
* collection of data completes successfully, EXCEPT if a time range
* end was specified on the command line.
* 
* History:
*   010809 x.y.z Louis <ID>: Created.
*   020121 0.0.13 louis <ID>: Implement explicit time range parameters
***********************************************************EndFunction*********/
int main(int argc, char *argv[])
{
  char *datadir, *rundir, *patterns[] = { "u*.log", "w*.log" }, **bufs = NULL;
  char *outfile, *propsfile, *progname;
  int set, nfiles, rundirsize, ext, oldnfiles, n_records = 0, nlines;
  FILE *fp[FILEBATCH], *outfp;
  unsigned int source, maxopenfiles, fileindex;
  time_t starttime = 0, stoptime = 0, exhaust_time = (LONG_MAX - 1),
    final_time = LONG_MAX, lasttime = 0;
  List files;
  struct tm bt;

  if (_NSTREAM_ < 6) {
    printf("The maximum allowed number of active file pointers is less than 6,\nand thus insufficient for the OPICS actuator.\n");
    exit(10);
  }
  if (argc < 2)
    usage();
  if (argv[1][0] == '-') {      /* manual operation */
    if (argc < 4)
      usage();
    debuglevel = 1;
    datadir = argv[2];
    outfile = argv[1] + 1;      /* skip the minus sign */
    propsfile = (argc > 4)? argv[4]: NULL;
    msgfp = *argv[3]? fopen(argv[3], "w"): NULL;
    maxopenfiles = (argc > 5)? atol(argv[5]): _NSTREAM_;
    if (maxopenfiles < 6 || maxopenfiles > _NSTREAM_)
      maxopenfiles = _NSTREAM_;
    if (argc > 6)
      starttime = (time_t) cal2time(argv[6]);
    if (argc > 7)
      stoptime = (time_t) cal2time(argv[7]);
  } else {                      /* standard actuator operation */
    if (argc < 13)
      usage();
    outfile = argv[1];
    if (*argv[7] == 'd')
      debuglevel = 1;
    datadir = argv[5];
    msgfp = *argv[6]? fopen(argv[6], "w"): NULL;
    propsfile = (argc > 13)? argv[13]: NULL;
    maxopenfiles = argv[8]? atol(argv[8]): _NSTREAM_;
    if (argc > 14)
      starttime = (time_t) cal2time(argv[14]);
    if (argc > 15)
      stoptime = (time_t) cal2time(argv[15]);
  }

  print_log("OPICS actuator; compiled " __DATE__ " " __TIME__ "\n");
  {
    time_t now = time(NULL);
    print_log("Current time: %s\n", asctime(localtime(&now)));
  }
  
  if (maxopenfiles < 6 || maxopenfiles > _NSTREAM_)
    maxopenfiles = _NSTREAM_;

  if (debuglevel)
    print_log("Starting (debug); maxopenfiles = %d\n", maxopenfiles);

  /* The OPICS logs consist of a great number of files, one per day per user
     ("u"-logs), plus one per day per workstation ("w"-logs), plus one per
     day ("b"-logs).  Even for fairly small set-ups these numbers may add
     up to more than the maximum supported number of simultaneously active file
     pointers.  Hence, we must carefully account for the number of file
     pointers in use.  These are:

     3 pointers for stdin, stdout, stderr
     1 pointer for output to a result file
     1 pointer for output to a log file

     The number of file pointers that we can use to simultaneously open
     OPICS log files is therefore _NSTREAM_ - 5

  */

  maxopenfiles -= 5;
  
  if (!starttime) {             /* wasn't in the command line arguments */
    if (propsfile && *propsfile != '\0') {
      /* get last-run-time */
      char timername[]="lastrun";
      char timeString[20]="0";
      int error = 0;
      Properties *props = CreateProperties(propsfile);
      
      if (debuglevel)
        print_log("Reading time of last collection from file %s... ", propsfile);
      error = getProperty(props,timername,timeString,20);
      if (!error) {
        starttime = (time_t) atoi(timeString);
        if (debuglevel)
          print_log("OK.\n");
      } else if (debuglevel)
        print_log("Failed.\n");

      DeleteProperties(props);
    } else
      print_log("No properties file was specified - collecting all data\n");
  }  
  if (!stoptime)            /* wasn't in the command line arguments */
    time(&stoptime);
  else
    propsfile = NULL;           /* so we don't update it at the end */
    
  if (debuglevel) {
    print_log("starttime = %d\n", starttime);
    print_log("stoptime  = %d\n", stoptime);
  }

  /* for use of findfirst() and cousins it is most convenient to be in the */
  /* directory where the files are.  We must also remember our present */
  /* location, in case the user requested an output file name relative to it. */
  rundir = getcwd(NULL, 512);
  if (!rundir) {
    print_log("Could not determine the current working directory\n");
    exit(2);
  } else if (debuglevel)
    print_log("current directory = %s\n", rundir);
  rundirsize = strlen(rundir);
  
  if (chdir(datadir)) {
    print_log("Could not change directory to %s\n", datadir);
    exit(3);
  }
  
  datadir = getcwd(NULL, 0);    /* user-supplied value may have been relative */
  if (debuglevel)
    print_log("Changed directory to %s\n", datadir);
  
  if (rundirsize + strlen(outfile) + 10 >= 512) { /* HACK: good for millions
                                                     of auxilliary files */
    print_log("Full output file name is too long:\n%s\\%s\n",
              rundir, outfile);
    exit(4);
  }
  
  /* we append the output file name to the run directory name */
  sprintf(rundir + rundirsize, "\\%s_", outfile);
  ext = strlen(rundir);
  sprintf(rundir + ext, "%1d", 0); /* first auxilliary file */

  /* unfortunately, the OPICS data is not always perfectly sorted in
     ascending temporal order, so we must go through each file, check,
     and sort if necessary.  We should not modify the existing files,
     because new records may still be appended to them while we're
     working on them, so we work with a list of files and modify that
     instead.  Also, some records are incomplete and do not end with a
     newline.  If a second ^B appears in a line then we discard the
     first (incomplete) record */
  if ((files = list_create(s_string, 0)) == NULL) {
    print_log("Could not create a list for file names\n");
    exit(5);
  }

  /* the OPICS log files are in subdirectories with names like 19981213 */
  {
    long handle;
    struct _finddata_t filedata;

    handle = _findfirst("*", &filedata);
    if (handle != -1L) {        /* we have some files */
      do
        if ((filedata.attrib & _A_SUBDIR)
            && (isdigit(filedata.name[0])))
        {
          int dirtime;
          char *p;

          dirtime = date2time(atol(filedata.name));
          if (dirtime < starttime - 86400) {
            print_log("Rejected directory %s because its name implies it is too old.\n",
                      filedata.name);
            continue;
          }
          chdir(filedata.name);
          if (debuglevel)
            print_log("Changed directory to %s\n", filedata.name);
          list_regular_files(2, patterns, files);
          chdir("..");
          if (debuglevel)
            print_log("Changed directory to ..\n");
        } else {
          print_log("Rejected %s because it is not a directory or does not have a date for name.\n", filedata.name);
        }
      while (!_findnext(handle, &filedata));
      _findclose(handle);
    }
  }

  {
    int i;

    i = rundir[rundirsize];
    rundir[rundirsize] = '\0';
    chdir(rundir);
    if (debuglevel)
      print_log("Changed directory to %s\n", rundir);
    rundir[rundirsize] = i;
  }
  
#ifdef OPICS6
  transform_files(files);       /* 040512 0.0.35 louis EE0404001: quick fix */
#endif
  
  sort_files(files, "_temp");
  nfiles = list_size(files);
  print_log("found %1d relevant u*.log and w*.log file(s)\n", nfiles);

  /* now <files> contains a list of <nfiles> names of files with data in
     ascending temporal order */

  set = 0;
  fileindex = 0;
  nlines = 0;
  while (fileindex < nfiles) {
    int i, line, done;
    long day, hour, minute, second;
    time_t time, times[FILEBATCH];
    char *p;
    unsigned int nfilesthisset = 0;
      
    do {
      char *filename;
      
      list_read(files, fileindex++, &filename);
      if (fp[nfilesthisset] = fopen(filename, "r"))
        nfilesthisset++;
    } while (nfilesthisset < maxopenfiles && fileindex < nfiles);
    
    /* we write the merged data into auxilliary files, because we may have */
    /* to make some higher-order merges if we have more than <maxopenfiles> */
    /* files to treat */
    sprintf(rundir + ext, "%1d", set);
    outfp = fopen(rundir, "w");
    if (!outfp) {
      print_log("Could not open file %s for writing\n", rundir);
      exit(6);
    }
    line = 0;                   /* number of output lines */

    /* we must read lines from all <nfilesthisset> files at the same time */
    /* to figure out what the temporal order must be */

    if (!bufs) { /* allocate memory for buffers for all open files */
      bufs = malloc((nfilesthisset + 1)*(sizeof(char *) + BUFSIZE));
      bufs[0] = (char *) bufs + (nfilesthisset + 1)*sizeof(char *);
      for (i = 1; i <= nfilesthisset; i++)
        bufs[i] =  bufs[i - 1] + BUFSIZE;
    }
    
    /* we must only treat data that falls between the temporal limits of */
    /* <starttime> and <stoptime>. */
    for (i = 0; i < nfilesthisset; i++) {
      do {
        if (fgets(bufs[i], BUFSIZE, fp[i])) {
          bt.tm_year = 70;
          bt.tm_mon = 0;
          bt.tm_mday = atol(bufs[i] + 10) - 25568; /* # days since 1970-1-1 */
          bt.tm_hour = atol(bufs[i] + 16);
          bt.tm_min = atol(bufs[i] + 19);
          bt.tm_sec = atol(bufs[i] + 22);
          bt.tm_isdst = -1;     /* let the system determine DST */
          times[i] = mktime(&bt);
        } else
          times[i] = exhaust_time;
      } while (times[i] < starttime);
      if (times[i] >= stoptime) {
        /* there wasn't any data in the target interval */
        fclose(fp[i]);
        times[i] = exhaust_time;
      }
    }
    
    initialize_rs_merge(nfilesthisset, &exhaust_time, &final_time, keycmp);
    
    source = UINT_MAX;
    for (i = 0; i < nfilesthisset; i++)
      source = rs_merge(i, &times[i]);

    while (source != UINT_MAX) {
      int donehere, gotnl;
      
      /* The input records arrive here in ascending temporal order. */
      /* The one to output right now came from file number <source>. */
      /* This file may already be closed, if none of its data falls */
      /* between the specified temporal boundaries. */
      /* We remove any 0x2 and prepend a date/time field. */
      p = strchr(bufs[source], '\x02');
      if (p) {                  /* a new record */
        char *pp;
        
        while (pp = strchr(p + 1, '\x02')) /* skip incomplete records */
                                /* TODO: accept incompletes */
                                /* if they're not too short */
          p = pp;
        if (line++)
          putc('\n', outfp);    /* finish previous record */
        /* a "w" file has whitespaces for the OPICS user ID */
        fprintf(outfp, "%1s %10d ", (p[107] == ' ')? "W": "U",
                times[source]);
        n_records++;
        p++;                  /* skip the 0x2 */
        while (*p) {
          if (isprint(*p))
            putc(*p, outfp);
          else if (*p == 29)      /* "DB Fields" command line separator */
            putc('&', outfp);     /* haven't seen & in example data, so we can
                                     (presumably) use it */
          else
            putc(' ', outfp);     /* generic non-printables */
          p++;
        }
        gotnl = (p[-1] == '\n');
      }

      /* Read the next line(s) from the file from which we just output data.
         Some lines end in 0x0 0xd 0xa, which translates (through fgets)
         into \n\n */
      do {
        if (fgets(bufs[source], BUFSIZE, fp[source])) {
          nlines++;
          if (bufs[source][0] == '\n') { /* a 0x0 0xd 0xa record end */
            donehere = 0;
            continue;           /* just get the next one */
          }
          p = strchr(bufs[source], '\x02');
          if (p) {              /* a new record */
            char *pp;
            
            while (pp = strchr(p + 1, '\x02'))
              p = pp;
            bt.tm_year = 70;
            bt.tm_mon = 0;
            bt.tm_mday = atol(p + 10) - 25568; /* # days since 1970-1-1 */
            bt.tm_hour = atol(p + 16);
            bt.tm_min = atol(p + 19);
            bt.tm_sec = atol(p + 22);
            bt.tm_isdst = -1;     /* let the system determine DST */
            times[source] = mktime(&bt);
            if (times[source] >= stoptime) {
              times[source] = exhaust_time;
              fclose(fp[source]);
            }
            source = rs_merge(source, &times[source]);
            donehere = 1;
          } else {  /* a second or further line from the previous */
                                /* record */
            if (gotnl)          /* this second line was caused by an explicit
                                 newline in the data, rather than by the limited
                                 size of our input buffer */
              putc(' ', outfp); /* we insert a whitespace */
            p = bufs[source];
            while (*p) {
              if (isprint(*p))
                putc(*p, outfp);
              p++;
            }
            donehere = 0;
          }
        } else {                /* this file is done */
          fclose(fp[source]);
          source = rs_merge(source, &exhaust_time);
          donehere = 1;
        } /* end of if (fgets(...)) else */
      } while (!donehere);
    } /* end of while (source != UINT_MAX) */
    rs_merge(0, NULL);          /* clean up nicely */
    if (line)
      putc('\n', outfp);          /* final newline */
    fclose(outfp);
    set++;
  } /* while (fileindex < nfiles) */

  /* we have now reduced the u and w files to one or a few merged files.
     We can now delete the temporary sorted files */
  {
    int i;
    char *filename;

    for (i = 0; i < nfiles; i++) {
      list_read(files, i, &filename);
      if (!strncmp(filename, "_temp_", 6)) {
        if (debuglevel)
          print_log("Removing file %s\n", filename);
        remove(filename);
      }
    }
    list_delete(files);
  }  

  /* we've treated the u and w files, and now it's the turn of the b files */
  {
    long handle, handle0;
    struct _finddata_t filedata;
    int daynum1, daynum2, daynum = 0, count = 0;
    char buf[BUFSIZE];
    FILE *fp, *outfp;
    int hour, minute, second;
    time_t time;

    daynum1 = starttime/86400 + 25567;
    daynum2 = stoptime/86400 + 25567;
    sprintf(rundir + ext, "%1d", set);
    outfp = fopen(rundir, "w");
    if (!outfp) {
      print_log("Could not open file %s for writing\n", rundir);
      exit(7);
    }
    
    /* there is no guarantee that _findnext() yields files in any particular
       order, so we must check each b*.bsy file individually */
    chdir(datadir);
    if (debuglevel)
      print_log("Changed directory to %s\n", datadir);
    handle0 = _findfirst("*", &filedata);
    if (handle0 != -1L) {
      do
        if ((filedata.attrib & _A_SUBDIR)
            && (isdigit(filedata.name[0])))
        {
          chdir(filedata.name);
          if (debuglevel)
            print_log("Changed directory to %s\n", filedata.name);
          handle = _findfirst("b*.bsy", &filedata);
          if (handle != -1L) {  /* we have some b*.bsy files */
            count = 0;
            daynum = atol(filedata.name + 1);
            do {
              if (daynum >= daynum1 && daynum <= daynum2) {
                fp = fopen(filedata.name, "r");
                count++;
                time = 0;
                while (fgets(buf, BUFSIZE, fp) && time < stoptime) {
                  int len;

                  len = strlen(buf);
                  bt.tm_year = 70;
                  bt.tm_mon = 0;
                  bt.tm_mday = daynum - 25568; /* # days since 1970-1-1 */
                  bt.tm_hour = atol(buf + 12);
                  bt.tm_min = atol(buf + 15);
                  bt.tm_sec = atol(buf + 18);
                  bt.tm_isdst = -1;     /* let the system determine DST */
                  time = mktime(&bt);
                  if (time >= starttime && time < stoptime) {
                    fprintf(outfp, "%1s %10d ", "B", time);
                    fputs(buf, outfp);
                    n_records++;
                  } /* end of if (time >= starttime ...) */
                  if (buf[len - 1] != '\n')
                    /* this record is larger than our buffer; read the rest */
                    do {
                      fgets(buf, BUFSIZE, fp);
                      fputs(buf, outfp);
                      len = strlen(buf);
                    } while (buf[len - 1] != '\n');
                } /* end of while (fgets()...) */
                fclose(fp);
              } /* end of if (daynum >= daynum1 ...) */
              if (!_findnext(handle, &filedata)) {
                daynum = atol(filedata.name + 1);
              } else
                daynum = INT_MAX;
            } while (daynum != INT_MAX);
            _findclose(handle);
          } /* end of if (handle != -1L) */
          chdir("..");
          if (debuglevel)
            print_log("Changed directory to ..\n");
        } /* end of if ((filedata.attrib ...)) */
      while (!_findnext(handle0, &filedata));
      _findclose(handle0);
    } /* end of if (handle0 != -1L) */
    fclose(outfp);
    set++;
    print_log("found %1d relevant b*.bsy files\n", count);
  }

  print_log("reduced to %1d records in %1d merged files\n", n_records, set);
  /* TODO: # records is inaccurate if # u,w files == 1: FIX
  
  /* We now have <set> auxilliary files that need to be merged; the data in */
  /* each is already in an acceptable form and in ascending temporal order. */
  if (set > 1) {                /* multiple merged files must themselves be */
                                /* merged */
    char *patterns[2];
    int direction = 0;
    time_t times[FILEBATCH];

    sprintf(rundir + rundirsize, "%s_*", outfile);
    patterns[0] = strdup(rundir + rundirsize);
    sprintf(rundir + rundirsize, "_%s*", outfile);
    patterns[1] = strdup(rundir + rundirsize);

    /* the sorted, merged files are in the run directory: go there  */
    rundir[rundirsize] = '\0';
    if (debuglevel)
      print_log("Changing directory to %s\n", rundir);
    chdir(rundir);

    do {
      set = 0;
      nfiles = 0;

      while (oldnfiles = nfiles,
             nfiles = open_many_files(1, patterns + direction, maxopenfiles,
                                      fp)) {
        int i;

        sprintf(rundir + rundirsize, direction? "\\%s_%1d": "\\_%s%1d",
                outfile, set);
        if (nfiles > 1) {
          /* we can use bufs[]; <nfiles> here cannot exceed the <nfilesthisset>
             for which room was allocated in bufs[] (except maybe if
             <maxopenfiles> is equal to 1, in which case we're in trouble
             anyway) */
          outfp = fopen(rundir, "w");
          if (!outfp) {
            print_log("Could not open file %s for writing\n", rundir);
            exit(8);
          }
      
          for (i = 0; i < nfiles; i++) {
            if (fgets(bufs[i], BUFSIZE, fp[i])) {
              times[i] = atol(bufs[i] + 2);
              if (times[i] > lasttime)
                lasttime = times[i];
            } else {
              fclose(fp[i]);
              times[i] = exhaust_time;
            }
          }
        
          initialize_rs_merge(nfiles, &exhaust_time, &final_time, keycmp);
          
          source = UINT_MAX;
          for (i = 0; i < nfiles; i++)
            source = rs_merge(i, &times[i]);

          while (source != UINT_MAX) {
            int len;
            
            fputs(bufs[source], outfp);
            /* the line we just output may not be a complete record; perhaps
               our buffer is too small.  We must check for this and properly
               treat overly long records. */
            len = strlen(bufs[source]);
            if (bufs[source][len - 1] != '\n')
              do {
                fgets(bufs[source], BUFSIZE, fp[source]);
                len = strlen(bufs[source]);
                fputs(bufs[source], outfp);
              } while (bufs[source][len - 1] != '\n' && len);
            if (fgets(bufs[source], BUFSIZE, fp[source])) {
              times[source] = atol(bufs[source] + 2);
              if (times[source] > lasttime)
                lasttime = times[source];
            } else {
              times[source] = exhaust_time;
              if (debuglevel)
                print_log("Closing file # %d\n", source);
              fclose(fp[source]);
            }
            source = rs_merge(source, &times[source]);
          }
          rs_merge(0, NULL);
          fclose(outfp);
        } else {                /* just one file left: copy */
          fclose(fp[0]);
          rs_merge(0, NULL);    /* cleanup */
          if (debuglevel)
            print_log("Renaming %s to %s\n", omf_lastfile(), rundir);
          rename(omf_lastfile(), rundir);
        }
        set++;
      } /* end of while (nfiles ...) */
      open_many_files(0, NULL, 0, NULL); /* clean up */

      /* we must now get rid of the old auxilliary files from the just */
      /* concluded pass */
      remove_files(patterns[direction]);

      print_log("reduced to %1d merged files\n", set);

      if (set > 1)
        direction = !direction;
    } while (set > 1);

    free(patterns[0]);
    free(patterns[1]);
  } /* end of if (set > 1) */
  
  if (set) {
    /* now, finally, we have a single merged file left, but it does not */
    /* yet have the correct name */
    char *auxfile;
    
    auxfile = strdup(rundir);
    sprintf(rundir + rundirsize, "\\%s", outfile);
    print_log("removing old %s (if it exists)\n", outfile);
    remove(rundir);
    print_log("renaming final merged file \"%s\" to \"%s\"\n", auxfile,
              rundir);
    rename(auxfile, rundir);
    free(auxfile);
  } else {                      /* no data was found; create an empty file */
    FILE *fp;

    print_log("no data found: creating an empty result file\n");
    fp = fopen(outfile, "w");
    fclose(fp);
  }
    
  if (debuglevel)
    print_size(outfile);

  if (propsfile && *propsfile != '\0') {
    char timername[]="lastrun";
    char timeString[20];
    int error = 0;
    Properties *props = CreateProperties(propsfile);

    /* 030911 0.0.31 louis: we've kept track of the time stamp of the
     * newest record that we've collected (in lasttime).  Next time,
     * we want to start collecting from just after that moment.  If we
     * did not collect any records this time, then we store the
     * starttime so we collect starting from the same time next
     * time. */
    sprintf(timeString, "%d", lasttime? lasttime + 1: starttime);

    setProperty(props, timername, timeString);

    DeleteProperties(props);
  }
  print_log("done\n");
  if (msgfp)
    fclose(msgfp);
  return 0;
}
