/**********************************************************BeginFile************
* Copyright:  (c) 2001 by Consul Risk Management B.V.  All Rights Reserved
* Time-stamp: <011004 Louis 14:22 rs_merge.c>
*
* Name:  replacement_selection.c
* Purpose:
*   Replacement Selection: Merging of data from multiple sources
* Notes:
*   After Algorithm R from Section 5.4.1 of Knuth's "The Art of Computer
*   Programming, Volume 3: Sorting and Searching", 2nd edition.
*
*   These routines can be applied in the following situation: you have
*   <p> (at least 2) sources of data, with the data in each in ascending
*   order (according to associated keys), and you want to merge those
*   sources together into one set in ascending order.
*
* History:
*   010806 <Version> louis <ID>: Created.
***********************************************************EndFile*************/
#include <malloc.h>             /* for malloc(), free() */
#include <limits.h>             /* for ULONG_MAX */

typedef unsigned int uint;

typedef struct node_struct {
  void *key;
  uint source;
  uint loser;
  uint rn;
  uint pe;
  uint pi;
} Node;

static Node *nodes;
static uint rmax, rc, rq, q, nactive, nsources;
static void *lastkey, *finalkey;
static int (*keycmp)(void *, void *);

/**********************************************************BeginFunction********
* Name: initialize_rs_merge
* Syntax:
*  int initialize_rs_merge(uint p, void *infinite_key,
*                          (int) (*cmp) (void *, void *))
* Parameters:
*   p            - (input) the number of sources of data; must be at least 2.
*   exhaust_key  - (input) a key value greater than any occurring in the data
*   infinite_key - (input) a key value greater than exhaust_key
*   cmp          - (input) a function that compares pointed-at keys and returns
*                  a negative value, zero, or a positive value, depending
*                  on whether the first key is less than, equal to, or
*                  greater than the second key
*
* Return Values: 0 = success; >0 = some error
* Purpose: Initialize for a replacement selection process.
*   
* History:
*   010806 <Version> louis <ID>: Created.
***********************************************************EndFunction*********/
int initialize_rs_merge(uint p, void *exhaust_key, void *infinite_key,
                         int (*cmp) (void *, void *))
{
  int i;

  nsources = p;
  
  if (p < 2) {
    nactive = 1;
    lastkey = infinite_key;
    finalkey = exhaust_key;
    keycmp = cmp;
    return 0;
  }

  /* step R1: Initialize */
  nodes = malloc(p*sizeof(Node));
  if (!nodes) {
    puts("Memory allocation failure in initialize_replacement_selection()");
    return 2;
  }
  rmax = 0;
  rc = 0;
  lastkey = infinite_key;
  finalkey = exhaust_key;
  
  q = 0;
  rq = 0;
  keycmp = cmp;
  nactive = p;                  /* we keep track of how many sources are */
                                /* still active, so we can return an */
                                /* appropriate stop value when all are done */

  for (i = 0; i < p; i++) {
    nodes[i].loser = i;
    nodes[i].rn = 0;
    nodes[i].pe = (i + p)/2;
    nodes[i].pi = i/2;
  }

  return 0;                     /* all OK */
}

/**********************************************************BeginFunction********
* Name: rs_merge
* Syntax:
*  uint rs_merge(uint source, void *key)
* Parameters:
*   uint source - (input) an identifier for the source of the data
*   void *key   - (input) the key associated with the next input record
*     <exhaust_key> (from intialize_rs_merge()) means that the <source>
*       has been exhausted.
*     NULL means to clean up (allocated memory and so on)
* Return Values: UINT_MAX if not all input slots are yet filled, or if the
*   current run is finished; otherwise the source identifier of the
*   smallest key currently in the system.
* Purpose: To merge input data from various sources.
* Notes:
*   Initialize with initialize_rs_merge().
*   You may use <exhaust_key> for <key> to indicate absence
*   of data for the associated source.  Once all sources are exhausted,
*   rs_merge() returns UINT_MAX.  To clean up, call rs_merge(0,NULL).
* History:
*   010806 <Version> louis <ID>: Created.
***********************************************************EndFunction*********/
uint rs_merge(uint source, void *key) {
  uint t;
  int notdone;

  if (nsources == 1) {
    if (key) {
      if (!keycmp(key,finalkey))
        return UINT_MAX;
      return 0;
    } else if (nactive > 0)
      return 0;
    else
      return UINT_MAX;
  }
  
  /* Step R4: Input new record */
  if (key) {                    /* input is not yet exhausted */
    if (!keycmp(key,finalkey))  /* user signals exhaustion of the source */
      nactive--;
    if (nactive <= 0)           /* no more active sources */
      rq = rmax + 1;
    else {
      nodes[q].source = source;
      nodes[q].key = key;
      if (keycmp(nodes[q].key,lastkey) < 0) {
        rq++;
        if (rq > rmax)
          rmax = rq;
      }
    }
  } else if (nactive > 0)       /* input is exhausted, but we still */
                                /* have active values in the system */
    rq = rmax + 1;
  else {                        /* we're done; clean up */
    free(nodes);
    return UINT_MAX;
  }

  /* Step R5: Prepare to update */
  t = nodes[q].pe;

  do {
    /* Step R6: Set new loser */
    if (nodes[t].rn < rq
        || ((nodes[t].rn == rq)
            && rq != 0          /* our own addition to algorithm R */
            && keycmp(nodes[nodes[t].loser].key, nodes[q].key) < 0)) {
      uint tempuint;
      
      tempuint = nodes[t].loser;
      nodes[t].loser = q;
      q = tempuint;
      
      tempuint = nodes[t].rn;
      nodes[t].rn = rq;
      rq = tempuint;
    }

    /* Step R7: Move up */
    if (t != 1) {
      t = nodes[t].pi;
      notdone = 1;
    } else
      notdone = 0;
  } while (notdone);
  
  /* Step R2: End of run? */
  if (rq != rc) {
    if (rq > rmax) {            /* we're done with this run */
      free(nodes);
      return UINT_MAX;
    } else
      rc = rq;
  }

  /* Step R3: Output top of tree */
  if (rq) {
    if (nactive) {
      lastkey = nodes[q].key;
      return nodes[q].source;
    } else                      /* our own stop value */
      return UINT_MAX;
  } else
    return UINT_MAX;
}
